/*
Copyright 2017 caicloud authors. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package gitlab

import (
	"crypto/tls"
	"encoding/json"
	"fmt"
	"io/ioutil"
	"net/http"
	"net/url"
	"strings"

	"github.com/caicloud/nirvana/log"
	v3 "gopkg.in/xanzy/go-gitlab.v0"

	c_v1alpha1 "github.com/caicloud/cyclone/pkg/apis/cyclone/v1alpha1"
	"github.com/caicloud/cyclone/pkg/server/apis/v1alpha1"
	"github.com/caicloud/cyclone/pkg/server/biz/scm"
	"github.com/caicloud/cyclone/pkg/util/cerr"
)

// V3 represents the SCM provider of API V3 GitLab.
type V3 struct {
	scmCfg *v1alpha1.SCMSource
	client *v3.Client
}

// GetToken gets the token by the username and password of SCM config.
func (g *V3) GetToken() (string, error) {
	return getOauthToken(g.scmCfg)
}

// CheckToken checks whether the token has the authority of repo by trying ListRepos with the token.
func (g *V3) CheckToken() error {
	if _, err := g.listReposInner(false); err != nil {
		return err
	}
	return nil
}

// ListRepos lists the repos by the SCM config.
func (g *V3) ListRepos() ([]scm.Repository, error) {
	return g.listReposInner(true)
}

// listReposInner lists the projects by the SCM config,
// list all projects while the parameter 'listAll' is true,
// otherwise, list projects by default 'ListPerPageOpt' number.
func (g *V3) listReposInner(listAll bool) ([]scm.Repository, error) {
	opt := &v3.ListProjectsOptions{
		ListOptions: v3.ListOptions{
			PerPage: scm.ListOptPerPage,
		},
	}

	// Get all pages of results.
	var allProjects []*v3.Project
	for {
		projects, resp, err := g.client.Projects.ListProjects(opt)
		if err != nil {
			return nil, convertGitlabError(err, resp)
		}

		allProjects = append(allProjects, projects...)
		if resp.NextPage == 0 || !listAll {
			break
		}
		opt.ListOptions.Page = resp.NextPage
	}

	repos := make([]scm.Repository, len(allProjects))
	for i, repo := range allProjects {
		repos[i].Name = repo.PathWithNamespace
		repos[i].URL = repo.HTTPURLToRepo
	}

	return repos, nil
}

// ListBranches lists the branches for specified repo.
// Only return first 20 branches, will not support listing all branches as v3 api has been deprecated.
func (g *V3) ListBranches(repo string) ([]string, error) {
	branches, resp, err := g.client.Branches.ListBranches(repo)
	if err != nil {
		log.Errorf("Fail to list branches for %s", repo)
		return nil, convertGitlabError(err, resp)
	}

	branchNames := make([]string, len(branches))
	for i, branch := range branches {
		branchNames[i] = branch.Name
	}

	return branchNames, nil
}

// ListTags lists the tags for specified repo.
// Only return first 20 tags, will not support listing all tags as v3 api has been deprecated.
func (g *V3) ListTags(repo string) ([]string, error) {
	tags, resp, err := g.client.Tags.ListTags(repo)
	if err != nil {
		log.Errorf("Fail to list tags for %s", repo)
		return nil, convertGitlabError(err, resp)
	}

	tagNames := make([]string, len(tags))
	for i, tag := range tags {
		tagNames[i] = tag.Name
	}

	return tagNames, nil
}

// ListPullRequests lists the merge requests for specified repo.
func (g *V3) ListPullRequests(repo, state string) ([]scm.PullRequest, error) {
	// GitLab mr state: opened, closed, locked, merged, all
	var s string
	switch state {
	case scm.PullRequestStateOpen:
		s = openedPullRequestState
	case openedPullRequestState, "closed", "locked", "merged", "all":
		s = state
	default:
		return nil, cerr.ErrorUnsupported.Error("GitLab(v3) pull request state", state)
	}

	opts := &v3.ListMergeRequestsOptions{
		State: &s,
		ListOptions: v3.ListOptions{
			PerPage: scm.ListOptPerPage,
		},
	}

	var allPRs []scm.PullRequest
	for {
		prs, resp, err := g.client.MergeRequests.ListMergeRequests(repo, opts)
		if err != nil {
			log.Errorf("Fail to list merge requests for %s", repo)
			return nil, convertGitlabError(err, resp)
		}

		for _, p := range prs {
			allPRs = append(allPRs, scm.PullRequest{
				ID:           p.IID,
				Title:        p.Title,
				Description:  p.Description,
				State:        p.State,
				TargetBranch: p.TargetBranch,
			})
		}
		if resp.NextPage == 0 {
			break
		}
		opts.Page = resp.NextPage
	}

	return allPRs, nil
}

// ListDockerfiles lists the Dockerfiles for specified repo.
func (g *V3) ListDockerfiles(repo string) ([]string, error) {
	// List Dockerfiles in a project with gitlab v3 api is very inefficient.
	// There is not a proper api can be used to do this with GitLab v3.
	//
	// FYI:
	// https://stackoverflow.com/questions/25127695/search-filenames-with-gitlab-api
	return nil, fmt.Errorf("list gitlab v3 dockerfiles not implemented")
}

// CreateStatus generate a new status for repository.
func (g *V3) CreateStatus(status c_v1alpha1.StatusPhase, targetURL, repo, commitSha string) error {
	state, description := transStatus(status)

	context := "continuous-integration/cyclone"
	opt := &v3.SetCommitStatusOptions{
		State:       v3.BuildState(state),
		Description: &description,
		TargetURL:   &targetURL,
		Context:     &context,
	}
	_, resp, err := g.client.Commits.SetCommitStatus(repo, commitSha, opt)
	return convertGitlabError(err, resp)
}

// GetPullRequestSHA gets latest commit SHA of pull request.
func (g *V3) GetPullRequestSHA(repo string, number int) (string, error) {
	path := fmt.Sprintf("%s/api/%s/projects/%s/merge_requests?iid=%d",
		strings.TrimSuffix(g.scmCfg.Server, "/"), v3APIVersion, url.QueryEscape(repo), number)
	req, err := http.NewRequest(http.MethodGet, path, nil)
	if err != nil {
		return "", err
	}

	if len(g.scmCfg.User) == 0 {
		req.Header.Set("PRIVATE-TOKEN", g.scmCfg.Token)
	} else {
		req.Header.Set("Authorization", "Bearer "+g.scmCfg.Token)
	}

	req.Header.Set("Content-Type", "application/json")
	httpClient := &http.Client{
		Transport: &http.Transport{
			TLSClientConfig: &tls.Config{InsecureSkipVerify: true},
		},
	}
	resp, err := httpClient.Do(req)
	if err != nil {
		log.Errorf("Fail to get project merge request as %s", err.Error())
		return "", err
	}

	defer func() {
		if err := resp.Body.Close(); err != nil {
			log.Errorf("Fail to close response body as: %v", err)
		}
	}()

	body, err := ioutil.ReadAll(resp.Body)
	if err != nil {
		log.Errorf("Fail to get project merge request as %s", err.Error())
		return "", err
	}

	if resp.StatusCode/100 == 2 {
		mr := []mergeRequestResponse{}
		err := json.Unmarshal(body, &mr)
		if err != nil {
			return "", err
		}
		if len(mr) > 0 {
			return mr[0].SHA, nil
		}
		return "", fmt.Errorf("Merge request %d not found ", number)
	}

	err = fmt.Errorf("Fail to get merge request %d as %s ", number, body)
	return "", err
}

// mergeRequestResponse represents the response of Gitlab merge request API.
type mergeRequestResponse struct {
	ID           int    `json:"id"`
	IID          int    `json:"iid"`
	TargetBranch string `json:"target_branch"`
	SHA          string `json:"sha"`
}

// GetWebhook gets webhook from specified repo.
func (g *V3) GetWebhook(repo string, webhookURL string) (*v3.ProjectHook, error) {
	hooks, resp, err := g.client.Projects.ListProjectHooks(repo, nil)
	if err != nil {
		return nil, convertGitlabError(err, resp)
	}

	for _, hook := range hooks {
		if hook.URL == webhookURL {
			return hook, nil
		}
	}

	return nil, cerr.ErrorContentNotFound.Error(fmt.Sprintf("webhook url %s", webhookURL))
}

// CreateWebhook creates webhook for specified repo.
func (g *V3) CreateWebhook(repo string, webhook *scm.Webhook) error {
	if webhook == nil || len(webhook.URL) == 0 || len(webhook.Events) == 0 {
		return fmt.Errorf("The webhook %v is not correct", webhook)
	}

	_, err := g.GetWebhook(repo, webhook.URL)
	if err != nil {
		if !cerr.ErrorContentNotFound.Derived(err) {
			return err
		}

		hook := generateV3ProjectHook(webhook)
		_, resp, err := g.client.Projects.AddProjectHook(repo, hook)
		if err != nil {
			return convertGitlabError(err, resp)
		}
		return nil
	}

	log.Warningf("Webhook already existed: %+v", webhook)
	return err
}

// DeleteWebhook deletes webhook from specified repo.
func (g *V3) DeleteWebhook(repo string, webhookURL string) error {
	hook, err := g.GetWebhook(repo, webhookURL)
	if err != nil {
		return err
	}

	if resp, err := g.client.Projects.DeleteProjectHook(repo, hook.ID); err != nil {
		log.Errorf("delete project hook %s for repo error: %v", hook.ID, repo, err)
		return convertGitlabError(err, resp)
	}

	return nil
}

func generateV3ProjectHook(webhook *scm.Webhook) *v3.AddProjectHookOptions {
	enableState, disableState := true, false
	// Push event is enable for Gitlab webhook in default, so need to remove this default option.
	hook := &v3.AddProjectHookOptions{
		PushEvents: &disableState,
	}

	for _, e := range webhook.Events {
		switch e {
		case scm.PullRequestEventType:
			hook.MergeRequestsEvents = &enableState
		case scm.PullRequestCommentEventType:
			hook.NoteEvents = &enableState
		case scm.PushEventType:
			hook.PushEvents = &enableState
		case scm.TagReleaseEventType:
			hook.TagPushEvents = &enableState
		default:
			log.Errorf("The event type %s is not supported, will be ignored", e)
			return nil
		}
	}
	hook.URL = &webhook.URL

	return hook
}
